'use strict';

const _Vue = require('vue');
const intlLocalematcher = require('@formatjs/intl-localematcher');
const ErrorEvent = require('./vue-intl-controller.6026ecb5.cjs');
const vue = require('./vue-intl-controller.27d34bfc.cjs');
const extendedIntl = require('@braw/extended-intl');
const asyncComputed = require('@braw/async-computed');

function newObject() {
  return Object.create(null);
}
function defineGetters(map) {
  const result = newObject();
  for (const property of [...Object.getOwnPropertyNames(map), ...Object.getOwnPropertySymbols(map)]) {
    Object.defineProperty(result, property, {
      configurable: true,
      get() {
        return map[property];
      }
    });
  }
  return result;
}
function defineRefGetters(referencesMap) {
  const result = ErrorEvent.createHashMap();
  for (const propertyKey of [...Object.getOwnPropertyNames(referencesMap), ...Object.getOwnPropertySymbols(referencesMap)]) {
    let normalizedKey = propertyKey;
    if (typeof normalizedKey === 'string' && normalizedKey.startsWith('$')) {
      normalizedKey = normalizedKey.slice(1);
    }
    Object.defineProperty(result, normalizedKey, {
      configurable: true,
      get: _Vue.isRef(referencesMap[propertyKey]) ? function referenceGetter() {
        return referencesMap[propertyKey].value;
      } : function getter() {
        return referencesMap[propertyKey];
      }
    });
  }
  return result;
}

// imagine typescript had variadic merge haha kidding... unless 😳
function mergeDescriptors(_source1, _source2, _source3, _source4, _source5, _source6, _source7, _source8, _source9) {
  const result = newObject();

  // eslint-disable-next-line prefer-rest-params
  for (const object of arguments) {
    if (object == null) continue;
    for (const property of [...Object.getOwnPropertyNames(object), ...Object.getOwnPropertySymbols(object)]) {
      Object.defineProperty(result, property, Object.getOwnPropertyDescriptor(object, property));
    }
  }
  return result;
}

function usePrefersPartial($config) {
  const $sources = _Vue.shallowRef([]);
  vue.observe(() => $config.preferredLocaleSources, updatedSources => {
    const entries = $sources.value;
    const newEntries = [];
    if (updatedSources != null) {
      for (const entry of entries) {
        if (!updatedSources.includes(entry.source)) {
          var _entry$source$uninsta, _entry$source;
          (_entry$source$uninsta = (_entry$source = entry.source).uninstall) === null || _entry$source$uninsta === void 0 ? void 0 : _entry$source$uninsta.call(_entry$source);
          entry.installed = false;
        }
      }
      for (const source of updatedSources) {
        const entry = entries.find(it => it.source === source);
        if (entry == null || !entry.installed) {
          var _source$install;
          (_source$install = source.install) === null || _source$install === void 0 ? void 0 : _source$install.call(source);
        }
        newEntries.push({
          installed: true,
          source
        });
      }
    }
    $sources.value = newEntries;
  });
  const $preferredLocale = _Vue.computed(() => {
    for (const entry of $sources.value) {
      const {
        source
      } = entry;
      let value;
      if (_Vue.isRef(source.prefers)) {
        value = source.prefers.value;
      } else {
        value = source.prefers;
      }
      if (value == null) {
        continue;
      }
      return intlLocalematcher.match(value, $config.locales.map(it => it.code), $config.defaultLocale);
    }
    return $config.defaultLocale;
  });
  function addSource(source) {
    let prepend = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
    const filteredSources = [];

    // unshift is quite costly operation, better use two ifs
    if (prepend) filteredSources.push(source);
    if ($config.preferredLocaleSources != null) {
      for (const existingSource of $config.preferredLocaleSources) {
        if (existingSource === source) continue;
        filteredSources.push(existingSource);
      }
    }
    if (!prepend) filteredSources.push(source);
    $config.preferredLocaleSources = filteredSources;
  }
  function removeSource(source) {
    const filteredSources = [];
    let shallUpdate = false;
    for (const existingSource of $config.preferredLocaleSources) {
      if (existingSource === source) {
        shallUpdate = true;
      } else {
        filteredSources.push(existingSource);
      }
    }
    if (shallUpdate) {
      $config.preferredLocaleSources = filteredSources;
    }
  }
  const refs = {
    $preferredLocale
  };
  return mergeDescriptors(defineRefGetters(refs), {
    addSource,
    removeSource
  });
}

/** Represents the controller configuration (state). */

function useConfigPartial(initialConfiguration) {
  const $config = _Vue.reactive((() => {
    var _initialConfiguration, _initialConfiguration2, _initialConfiguration3, _initialConfiguration4;
    const defaultLocale = (_initialConfiguration = initialConfiguration === null || initialConfiguration === void 0 ? void 0 : initialConfiguration.defaultLocale) !== null && _initialConfiguration !== void 0 ? _initialConfiguration : 'en-US';
    const locale = (_initialConfiguration2 = initialConfiguration === null || initialConfiguration === void 0 ? void 0 : initialConfiguration.locale) !== null && _initialConfiguration2 !== void 0 ? _initialConfiguration2 : defaultLocale;
    let locales = initialConfiguration === null || initialConfiguration === void 0 ? void 0 : initialConfiguration.locales;
    if (locales == null) {
      locales = [];
      locales.push({
        code: defaultLocale
      });
      if (locale !== defaultLocale) {
        locales.push({
          code: locale
        });
      }
    }
    return {
      defaultLocale,
      locale,
      locales,
      usePreferredLocale: (_initialConfiguration3 = initialConfiguration === null || initialConfiguration === void 0 ? void 0 : initialConfiguration.usePreferredLocale) !== null && _initialConfiguration3 !== void 0 ? _initialConfiguration3 : false,
      preferredLocaleSources: (_initialConfiguration4 = initialConfiguration === null || initialConfiguration === void 0 ? void 0 : initialConfiguration.preferredLocaleSources) !== null && _initialConfiguration4 !== void 0 ? _initialConfiguration4 : []
    };
  })());
  const $intlLocale = _Vue.computed(() => new Intl.Locale($config.locale));
  return mergeDescriptors(defineGetters({
    $config
  }), {
    get defaultLocale() {
      return $config.defaultLocale;
    }
  }, defineRefGetters({
    $intlLocale
  }));
}

function useLocaleDataPartial($config, $locales) {
  function getLocale(localeCode) {
    const locale = $locales.value[localeCode];
    if (locale == null) {
      throw new Error(`Unknown locale "${localeCode}"`);
    }
    return locale;
  }
  const $defaultMessages = _Vue.computed(() => getLocale($config.defaultLocale).messages);
  const $messages = _Vue.computed(() => getLocale($config.locale).messages);
  const $defaultResources = _Vue.computed(() => getLocale($config.defaultLocale).resources);
  const $resources = _Vue.computed(() => getLocale($config.locale).resources);
  const refs = {
    $defaultMessages,
    $messages,
    $defaultResources,
    $resources
  };
  return mergeDescriptors(defineGetters(refs), defineRefGetters(refs));
}

function useEventTargetPartial($controller) {
  const eventRegister = new Map();

  // @ts-expect-error There appears to be a bug in TS that makes it think this type is incompatible.

  function addEventListener(type, listener, options) {
    if (!eventRegister.has(type)) {
      eventRegister.set(type, new Set());
    }
    const entries = eventRegister.get(type);
    for (const entry of entries) {
      if (entry.listener === listener) {
        entries.delete(entry);
      }
    }
    entries.add({
      listener,
      options: {
        ...options
      }
    });
  }

  // @ts-expect-error See line above addEventListener.

  function removeEventListener(type, listener) {
    const entries = eventRegister.get(type);
    if (entries == null) return;
    for (const value of entries) {
      if (value.listener === listener) {
        entries.delete(value);
      }
    }
  }
  function dispatchEvent(event) {
    const eventType = event.type;
    const entries = eventRegister.get(eventType);
    let canceled = false;
    const allowAsync = ErrorEvent.isAsyncEvent(event);

    // If you remember that async functions in JavaScript are essentially state
    // machines, this should not scare you.
    //
    // Basically, unless we are dealing with async event, this function will
    // never go beyond state 0, therefore it will act like a regular function
    // but with extra steps. Which is okay-ish compromise that helps us to avoid
    // code repetition.

    async function impl() {
      if (entries == null) return;
      const sortedListeners = [...entries.values()].sort((a, b) => {
        var _b$options$priority, _a$options$priority;
        return ((_b$options$priority = b.options.priority) !== null && _b$options$priority !== void 0 ? _b$options$priority : 0) - ((_a$options$priority = a.options.priority) !== null && _a$options$priority !== void 0 ? _a$options$priority : 0);
      });
      for (const entry of sortedListeners) {
        var _entry$options$always, _entry$options$once;
        if (canceled && !((_entry$options$always = entry.options.always) !== null && _entry$options$always !== void 0 ? _entry$options$always : false)) {
          continue;
        }
        try {
          const ret = entry.listener.call($controller.value, event);
          if (allowAsync) await ret;
        } catch (err) {
          var _eventRegister$get$si, _eventRegister$get;
          if (((_eventRegister$get$si = (_eventRegister$get = eventRegister.get('error')) === null || _eventRegister$get === void 0 ? void 0 : _eventRegister$get.size) !== null && _eventRegister$get$si !== void 0 ? _eventRegister$get$si : 0) < 1) {
            // eslint-disable-next-line no-console
            console.error('Uncaught', err);
          } else {
            const wrappedError = new Error(`An error occurred while calling the event listener for "${eventType}"`, {
              cause: err
            });
            dispatchEvent(new ErrorEvent.ErrorEvent(wrappedError));
          }
        }
        if ((_entry$options$once = entry.options.once) !== null && _entry$options$once !== void 0 ? _entry$options$once : false) {
          entries.delete(entry);
        }
        if (!canceled) canceled = ErrorEvent.isCanceled(event);
      }
    }
    const completion = impl();
    if (entries != null && entries.size === 0) {
      eventRegister.delete(event.type);
    }
    if (allowAsync) {
      return completion.then(() => !canceled, () => false);
    } else {
      return !canceled;
    }
  }
  return {
    addEventListener,
    removeEventListener,
    dispatchEvent
  };
}

const formatAliases = {
  formatCompactNumber: 'compactNumber',
  formatCustomMessage: 'customMessage',
  formatDate: 'date',
  formatDateTimeRange: 'dateTimeRange',
  formatDateToParts: 'dateToParts',
  formatDisplayName: 'displayName',
  formatList: 'list',
  formatListToParts: 'listToParts',
  formatMessage: 'message',
  formatNumber: 'number',
  formatNumberToParts: 'numberToParts',
  formatPlural: 'plural',
  formatRelativeTime: 'relativeTime',
  formatTime: 'time',
  formatTimeToParts: 'timeToParts',
  $ago: '$ago',
  formatTimeDifference: 'timeDifference'
};
// ~~~~~~~
// NOTE: You might need to remove as const so auto-complete works, also you can't place it after satisfier, that's not supported :(

function useIntlPartial($config, $messages) {
  const $formats = _Vue.ref(ErrorEvent.createHashMap());
  const intlCache = extendedIntl.createIntlCache();
  const $intl = _Vue.computed(() => extendedIntl.createIntl({
    locale: $config.locale,
    defaultLocale: $config.defaultLocale,
    messages: $messages.value
  }, intlCache));
  vue.observe($intl, intl => {
    const formats = $formats.value;
    for (const [intlProp, alias] of Object.entries(formatAliases)) {
      formats[alias] = intl[intlProp];
    }
  });
  return defineRefGetters({
    $formats,
    $intl
  });
}

function useLocalesPartial(initialLocaleData, $config, eventTarget, prefersPartial) {
  const $locales = _Vue.ref(Object.assign(ErrorEvent.createHashMap(), initialLocaleData));
  vue.observe(() => $config.locales, availableLocales => {
    const knownLocaleCodes = new Set();
    const locales = $locales.value;
    for (const descriptor of availableLocales) {
      if (knownLocaleCodes.has(descriptor.code)) {
        throw new Error('Duplicate locale code detected');
      }
      knownLocaleCodes.add(descriptor.code);
      if (!ErrorEvent.hasKey(locales, descriptor.code)) {
        locales[descriptor.code] = {
          get descriptor() {
            return descriptor;
          },
          messages: ErrorEvent.createHashMap(),
          resources: ErrorEvent.createHashMap()
        };
      }
    }
    for (const code of Object.keys(locales)) {
      if (!knownLocaleCodes.has(code)) {
        delete locales[code];
      }
    }
  });
  const $automatic = _Vue.computed(() => $config.usePreferredLocale);
  const $automaticLocaleCode = _Vue.ref(null);
  const $locale = _Vue.computed(() => {
    var _$automaticLocaleCode;
    return (_$automaticLocaleCode = $automaticLocaleCode.value) !== null && _$automaticLocaleCode !== void 0 ? _$automaticLocaleCode : $config.locale;
  });
  const $loading = asyncComputed.asyncComputed({
    watch() {
      return $locales.value[$locale.value];
    },
    async get(locale) {
      if (locale == null) {
        throw new Error('No locale descriptor exists for the current locale');
      }
      const abortController = new AbortController();
      const event = new ErrorEvent.LocaleLoadEvent(locale);
      abortController.signal.addEventListener('abort', event.cancel.bind(null));
      if (!(await eventTarget.dispatchEvent(event))) {
        throw new Error(`Cannot load locale messages for the locale "${locale.descriptor.code}": load event is cancelled`, {
          cause: abortController.signal.reason
        });
      }
      Object.assign(locale, event.collect());
    }
  });
  vue.observe(() => ({
    isEnabled: $automatic.value,
    locale: prefersPartial.preferredLocale
  }), (_ref, previousState) => {
    let {
      isEnabled,
      locale
    } = _ref;
    const wasEnabled = previousState === null || previousState === void 0 ? void 0 : previousState.isEnabled;
    if (isEnabled && !wasEnabled) {
      var _previousState$locale;
      const previousLocale = (_previousState$locale = previousState === null || previousState === void 0 ? void 0 : previousState.locale) !== null && _previousState$locale !== void 0 ? _previousState$locale : $config.locale;
      const languageSwitchCanceled = !eventTarget.dispatchEvent(new ErrorEvent.LocaleChangeEvent(previousLocale, locale, true));
      if (languageSwitchCanceled) return;
      $automaticLocaleCode.value = locale;
    } else if (!isEnabled && wasEnabled) {
      $automaticLocaleCode.value = null;
    }
  });
  function getLocaleDescriptor(localeCode) {
    var _$locales$value$local;
    return (_$locales$value$local = $locales.value[localeCode]) === null || _$locales$value$local === void 0 ? void 0 : _$locales$value$local.descriptor;
  }
  function addLocale(descriptor, force) {
    const normalizedDescriptor = typeof descriptor === 'string' ? {
      code: descriptor
    } : descriptor;
    const newLocales = [];
    for (const locale of $config.locales) {
      if (locale.code === normalizedDescriptor.code) {
        if (!force) {
          throw new Error(`Locale "${normalizedDescriptor.code}" already exists`);
        }
      } else {
        newLocales.push(locale);
      }
    }
    newLocales.push(normalizedDescriptor);
    $config.locales = newLocales;
    return normalizedDescriptor;
  }
  function removeLocale(descriptor) {
    var _$config$locales$spli;
    let index;
    if (typeof descriptor === 'string') {
      index = $config.locales.findIndex(locale => locale.code === descriptor);
    } else {
      index = $config.locales.findIndex(locale => locale === descriptor);
    }
    return index === -1 ? null : (_$config$locales$spli = $config.locales.splice(index, 1)[0]) !== null && _$config$locales$spli !== void 0 ? _$config$locales$spli : null;
  }
  function addMessages(descriptor, messages) {
    let locale;
    if (typeof descriptor === 'string') {
      locale = $locales.value[descriptor];
    } else {
      for (const knownLocale of Object.values($locales.value)) {
        if (knownLocale.descriptor === descriptor) {
          locale = knownLocale;
        }
      }
    }
    if (locale == null) {
      if (typeof descriptor === 'string') {
        // eslint-disable-next-line unicorn/prefer-type-error
        throw new Error(`Locale "${descriptor}" does not exist`);
      } else {
        // eslint-disable-next-line unicorn/prefer-type-error
        throw new Error(`Locale with provided descriptor (for ${descriptor.code}) does not exist`);
      }
    }
    if (locale.messages == null) {
      locale.messages = ErrorEvent.createHashMap();
    }
    Object.assign(locale.messages, messages);
  }
  async function changeLocale(localeCode) {
    if ($config.locale === localeCode) {
      return;
    }
    if (localeCode === 'auto') {
      if (!eventTarget.dispatchEvent(new ErrorEvent.AutomaticStateChangeEvent(true))) {
        throw new Error('Enabling of automatic mode has been cancelled');
      }
    } else {
      const autoSwitchOffCanceled = eventTarget.dispatchEvent(new ErrorEvent.AutomaticStateChangeEvent(false));
      if ($automatic.value && !autoSwitchOffCanceled) {
        return;
      }
      const localeChangeCanceled = !eventTarget.dispatchEvent(new ErrorEvent.LocaleChangeEvent($config.locale, localeCode, false));
      if (localeChangeCanceled) {
        throw new Error(`Locale change to "${localeCode}" was cancelled`);
      }
      $config.locale = localeCode;
    }
    $config.usePreferredLocale = localeCode === 'auto';
    await $loading.promise;
  }
  return mergeDescriptors(defineGetters({
    $loading,
    $locales
  }), defineRefGetters({
    $automatic,
    $locale
  }), {
    get ready() {
      return $loading.fulfilled;
    },
    getLocaleDescriptor,
    addLocale,
    removeLocale,
    addMessages,
    changeLocale
  });
}

function createController(initialConfiguration, initialLocaleData) {
  const $controller = _Vue.ref(null);
  const configPartial = useConfigPartial(initialConfiguration);
  const eventTargetPartial = useEventTargetPartial($controller);
  const prefersPartial = usePrefersPartial(configPartial.$config);
  const localesPartial = useLocalesPartial(initialLocaleData, configPartial.$config, eventTargetPartial, prefersPartial);
  const dataPartial = useLocaleDataPartial(configPartial.$config, localesPartial.$locales);
  const intlPartial = useIntlPartial(configPartial.$config, dataPartial.$messages);
  $controller.value = mergeDescriptors(configPartial, localesPartial, intlPartial, dataPartial, prefersPartial, eventTargetPartial);
  return $controller.value;
}

exports.createController = createController;
